#include <message_filters/subscriber.h>
#include <message_filters/sync_policies/approximate_time.h>
#include <message_filters/synchronizer.h>
#include <ros/ros.h>
#include <pcl_ros/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl_ros/transforms.h>
#include <sensor_msgs/Imu.h>
#include <nav_msgs/GridCells.h>
#include <tf2_eigen/tf2_eigen.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

//External libraries
#include <Eigen/Core>
#include <Eigen/Geometry>

ros::Publisher map_pub;
int device_width, device_height, cam_width, cam_height;

typedef pcl::PointCloud<pcl::PointXYZ> PointCloud;

void conversion_callback(const PointCloud::ConstPtr& pcl, const sensor_msgs::Imu::ConstPtr& imu) {
  PointCloud pcl_ground;
  nav_msgs::GridCells obstacle_map;
  geometry_msgs::TransformStamped tf_camera_ground;
  Eigen::Affine3d ground_camera;
  Eigen::Vector3d ground;
  double roll, pitch;

  // Understanding the orientation of the headset camera,
  // transforming the pointcloud so that Z is pointing "upwards",
  // so we can create the density map of obstacles
  ground(0) = imu->linear_acceleration.x;
  ground(1) = imu->linear_acceleration.y;
  ground(2) = imu->linear_acceleration.z;
  ground.normalize();

  // Solving inverse kinematics as roll-pitch-yaw rotations
  // Yaw is not relevant
  roll = atan2(ground(1), ground(0));
  pitch = atan2(ground(1)*sin(roll), ground(2));

  // Calculate the transformation from the ground to the camera frame
  ground_camera = Eigen::AngleAxis<double>(roll, Eigen::Vector3d::UnitX())*
                     Eigen::AngleAxis<double>(pitch, Eigen::Vector3d::UnitY());

  tf_camera_ground = tf2::eigenToTransform(ground_camera.inverse());

  // Transforming the pointcloud for upwards orientation
  pcl_ros::transformPointCloud(*pcl, pcl_ground, tf_camera_ground.transform);

  // Initializing the obstacle map message
  // This message contains a map that describes
  // the obstacles as seen from the top of the room.
  // Higher values at each cell indicate higher chances of an obstacle
  obstacle_map.header.frame_id = "ground";
  obstacle_map.cell_width = cam_width/device_width;
  obstacle_map.cell_height = cam_height/device_height;

  // Iterating through each point of the point cloud, and assigning it
  // in a cell of the obstacle map. Each cell will correspond to one
  // element on the grid of the tactile device.
  BOOST_FOREACH (const pcl::PointXYZ& pt, pcl_ground.points) {
    // Figuring out which cell does each element correspond to.
    // i is the column, j is the row of the cell of the tactile device
    int i = pt.x / cam_width * device_width;
    int j = pt.y / cam_height * device_height;
    int cell_ind = i * device_width + j; // Calculating the index of the cell
    obstacle_map.cells[cell_ind].x = i * device_width;
    obstacle_map.cells[cell_ind].y = i * device_height;
    obstacle_map.cells[cell_ind].z += pt.z; // Accumulating the z values of all the points that fall in this cell
  }

  //Publishing the obstacle map for consumption from the tactile device
  map_pub.publish(obstacle_map);
}

int main(int argc, char** argv) {
  ros::init(argc, argv, "depth_to_2d");
  ros::NodeHandle nh;

  // Initializing the publisher of the obstacle map
  map_pub = nh.advertise<nav_msgs::GridCells>("/obstacle_map", 1);

  // Initialising the spatial resolution of the camera and the tactile device
  // These value are assumptions but can be updated to reflect real values
  cam_width = 1024;
  cam_height = 1024;
  device_width = 400;
  device_height = 400;

  // Creating subscribers to individual topics
  message_filters::Subscriber<PointCloud> pointcloud_sub(nh, "/depth_camera/pointcloud", 1);
  message_filters::Subscriber<sensor_msgs::Imu> imu_sub(nh, "/imu", 1);

  // Creating a filter that calls the callback only when messages arrive at approximately
  // the same time at each topic
  typedef message_filters::sync_policies::ApproximateTime<PointCloud, sensor_msgs::Imu> ApproxSyncPolicy;
  message_filters::Synchronizer<ApproxSyncPolicy> sync(ApproxSyncPolicy(1), pointcloud_sub, imu_sub);
  sync.registerCallback(conversion_callback);

  ros::spin();

  return 0;
}
